package errors_test

import (
	"errors"
	"fmt"
	"io/fs"
	"testing"
)

func TestIs(t *testing.T) {
	err1 := errors.New("1")

	erra := wrapped{"wrap 2", err1}

	errb := wrapped{"wrap 3", erra}

	err3 := errors.New("3")

	poser := &poser{"either 1 or 3", func(err error) bool {
		return err == err1 || err == err3
	}}

	testCases := []struct {
		err error

		target error

		match bool
	}{
		{nil, nil, true},

		{err1, nil, false},

		{err1, err1, true},

		{erra, err1, true},

		{errb, err1, true},

		{err1, err3, false},

		{erra, err3, false},

		{errb, err3, false},

		{poser, err1, true},

		{poser, err3, true},

		{poser, erra, false},

		{poser, errb, false},

		{errorUncomparable{}, errorUncomparable{}, true},

		{errorUncomparable{}, &errorUncomparable{}, false},

		{&errorUncomparable{}, errorUncomparable{}, true},

		{&errorUncomparable{}, &errorUncomparable{}, false},

		{errorUncomparable{}, err1, false},

		{&errorUncomparable{}, err1, false},
	}

	for _, tc := range testCases {
		t.Run("", func(t *testing.T) {
			if got := errors.Is(tc.err, tc.target); got != tc.match {
				t.Errorf("Is(%v, %v) = %v, want %v", tc.err, tc.target, got, tc.match)
			}
		})
	}
}

type poser struct {
	msg string

	f func(error) bool
}

var poserPathErr = &fs.PathError{Op: "poser"}

func (p *poser) Error() string { return p.msg }

func (p *poser) Is(err error) bool { return p.f(err) }

func (p *poser) As(err interface{}) bool {
	switch x := err.(type) {

	case **poser:

		*x = p

	case *errorT:

		*x = errorT{"poser"}

	case **fs.PathError:

		*x = poserPathErr

	default:

		return false

	}

	return true
}

type errorT struct{ s string }

func (e errorT) Error() string { return fmt.Sprintf("errorT(%s)", e.s) }

type wrapped struct {
	msg string

	err error
}

func (e wrapped) Error() string { return e.msg }

func (e wrapped) Unwrap() error { return e.err }

type errorUncomparable struct {
	f []string
}

func (errorUncomparable) Error() string {
	return "uncomparable error"
}

func (errorUncomparable) Is(target error) bool {
	_, ok := target.(errorUncomparable)

	return ok
}
